#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;

namespace HeuristicLab.Optimization.Networks {
  [Item("AlgorithmServiceNode", "A node of an optimization network which provides a HeuristicLab algorithm as a service.")]
  [StorableClass]
  public class AlgorithmServiceNode : Node, IAlgorithmServiceNode {
    private object locker = new object();
    private Dictionary<IAlgorithm, AutoResetEvent> waitHandles = new Dictionary<IAlgorithm,AutoResetEvent>();
    private Dictionary<IAlgorithm, Exception> exceptions = new Dictionary<IAlgorithm, Exception>();

    new public PortCollection Ports {
      get { return base.Ports; }
    }

    [Storable]
    private IAlgorithm template;
    public IAlgorithm Template {
      get { return template; }
      set {
        if (value != template) {
          template = value;
          OnTemplateChanged();
        }
      }
    }

    [Storable]
    private RunCollection runs;
    public RunCollection Runs {
      get { return runs; }
    }

    [StorableConstructor]
    protected AlgorithmServiceNode(bool deserializing) : base(deserializing) { }
    protected AlgorithmServiceNode(AlgorithmServiceNode original, Cloner cloner)
      : base(original, cloner) {
      template = cloner.Clone(original.template);
      runs = cloner.Clone(original.runs);
    }
    public AlgorithmServiceNode()
      : base("AlgorithmServiceNode") {
      runs = new RunCollection();
    }
    public AlgorithmServiceNode(string name)
      : base(name) {
      runs = new RunCollection();
    }
    public AlgorithmServiceNode(string name, string description)
      : base(name, description) {
      runs = new RunCollection();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new AlgorithmServiceNode(this, cloner);
    }

    protected virtual void UpdateParameters(IMessage message, CancellationToken token) {
      // set template parameters
      lock (locker) {
        if (template != null) {
          foreach (var v in message.Values) {
            IParameter param = null;
            if (Template.Parameters.TryGetValue(v.Name, out param)) {
              var vp = param as IValueParameter;
              if (vp != null) vp.Value = v.Value;
            }
            if (Template.Problem.Parameters.TryGetValue(v.Name, out param)) {
              var vp = param as IValueParameter;
              if (vp != null) vp.Value = v.Value;
            }
          }
        }
      }
    }

    protected virtual void Execute(IMessage message, CancellationToken token) {
      if (Template == null) throw new InvalidOperationException("Template is null");

      IAlgorithm algorithm;
      lock (locker) {
        algorithm = (IAlgorithm)Template.Clone();
        waitHandles.Add(algorithm, new AutoResetEvent(false));
      }

      // set parameters
      foreach (var v in message.Values) {
        IParameter param = null;
        if (algorithm.Parameters.TryGetValue(v.Name, out param)) {
          var vp = param as IValueParameter;
          if (vp != null) vp.Value = v.Value;
        }
        if (algorithm.Problem.Parameters.TryGetValue(v.Name, out param)) {
          var vp = param as IValueParameter;
          if (vp != null) vp.Value = v.Value;
        }
      }

      algorithm.StoreAlgorithmInEachRun = false;
      algorithm.Runs.Clear();
      algorithm.Prepare();
      algorithm.ExceptionOccurred += Algorithm_ExceptionOccurred;
      algorithm.Stopped += Algorithm_Stopped;

      using (token.Register(() => { algorithm.Stop(); })) {
        algorithm.Start();
        waitHandles[algorithm].WaitOne();
      }

      lock (locker) {
        waitHandles[algorithm].Dispose();
        waitHandles.Remove(algorithm);

        Exception ex = null;
        if (exceptions.TryGetValue(algorithm, out ex)) {
          exceptions.Remove(algorithm);
          throw ex;
        }
      }

      // retrieve results
      var run = algorithm.Runs.First();
      foreach (var v in message.Values) {
        IItem result = null;
        if (run.Results.TryGetValue(v.Name, out result)) {
          v.Value = result;
        }
      }

      lock (locker) {
        Runs.Add(run);
      }
    }

    private void Algorithm_ExceptionOccurred(object sender, EventArgs<Exception> e) {
      var algorithm = (IAlgorithm)sender;
      lock (locker) {
        exceptions.Add(algorithm, e.Value);
      }
      algorithm.Stop();
    }
    private void Algorithm_Stopped(object sender, EventArgs e) {
      lock (locker) {
        waitHandles[(IAlgorithm)sender].Set();
      }
    }

    public event EventHandler TemplateChanged;
    protected virtual void OnTemplateChanged() {
      var handler = TemplateChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    #region Ports Events
    protected override void RegisterPortsEvents() {
      base.RegisterPortsEvents();
      foreach (var p in Ports)
        RegisterPortEvents(p);
    }
    protected override void Ports_ItemsAdded(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_ItemsAdded(sender, e);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    protected override void Ports_ItemsRemoved(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_ItemsRemoved(sender, e);
      foreach (var p in e.Items)
        DeregisterPortEvents(p);
    }
    protected override void Ports_ItemsReplaced(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_ItemsReplaced(sender, e);
      foreach (var p in e.OldItems)
        DeregisterPortEvents(p);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    protected override void Ports_CollectionReset(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_CollectionReset(sender, e);
      foreach (var p in e.OldItems)
        DeregisterPortEvents(p);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    #endregion

    #region Port Events
    private void RegisterPortEvents(IPort port) {
      IGenericPort g = port as IGenericPort;
      if (g != null) {
        g.MessageReceived += GenericPort_MessageReceived;
      }
    }
    private void DeregisterPortEvents(IPort port) {
      IGenericPort g = port as IGenericPort;
      if (g != null) {
        g.MessageReceived -= GenericPort_MessageReceived;
      }
    }
    private void GenericPort_MessageReceived(object sender, EventArgs<IMessage, CancellationToken> e) {
      var port = (IGenericPort)sender;
      if (port.Parameters.Any(p => p.Type.HasFlag(PortParameterType.Output)))
        Execute(e.Value, e.Value2);  // execute algorithm to compute output
      else
        UpdateParameters(e.Value, e.Value2);  // update template parameters
    }
    #endregion
  }
}
